/*
 * Modification History
 *
 * 2000-November-9		Jason Rohrer
 * Created.
 *
 * 2000-November-12		Jason Rohrer
 * Fixed bug in the way that ranking numbers are calculated.
 *
 * 2000-November-14		Jason Rohrer
 * Changed so that agent exceptions that occur in playRound are
 * cocumented an affect scoring.
 *
 * 2001-June-2   Jason Rohrer
 * Changed to use a version of the StreamTokenizer constructor
 * that is not deprecated.
 * Added a statement to close the file stream.
 *
 * 2001-June-3   Jason Rohrer
 * Changed to use a specifiable set of standard agents.
 * Changed to read settings from an INI file.
 * Changed to ignore comments in input files.
 * Added line number printouts for input file errors.
 * Fixed a bug in the board sets directory listing.
 * Fixed location of ranking file.
 */

package ctf.environment;

import ctf.agent.*;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.*;
import java.awt.event.*;

import java.awt.Image;

import java.io.*;
import java.util.Vector;
/**
 * Tournament for CTF.
 *
 * @author Jason Rohrer
 */
public class Tournament implements StateChangeListener {

	
	int mExceptionPenalty = 5;

	// the fraction of maps a student agent must win on against
	// a standard agent to get full credit against that agent
	float mMapFractionToWin = 0.5f;

	int mCorrectAgentPoints = 60;

	// the number of points awarded to the winner of the
	// student agent tournament
	// points for other student agents in the tournament
	// scale linearly down from this value
	int mTournamentPoints = 10;

	
	int mTeam1Score = 0;
	int mTeam2Score = 0; 
	
	int mDrawThreshold = 200;

	String mIniFileName = "Tournament.ini";
	String mSetsDirName = "sets";
	String mStandardAgentClassesFileName = "StandardAgentClasses.txt";
	String mCompetitorAgentClassesFileName = "CompetitorAgentClasses.txt";
	String mStatsDirName = "stats";
		
	Game mGame;
	FileSet mBoardSets[];
	String mBoardSetNames[];
	//ObstacleSet mObstacleSet;
	
	Class mAvailableCompetitorClasses[];
	int mCompetitorWins[];	// count of how many wins each class has garnered
	
	Class mStandardAgentClasses[];
	// the maximum number of points granted for
	// good performance against a standard agent
	int mPointsPerStandardAgent[];	
	//Class mJcr13AgentClass;
	//Class mRandomAgentClass;
	//Class mSecretAgentClass;
	
	// mWinsAgainstMainAgents[a][i] = # of games agent _a_ won against
	// agent class mStandardAgentClasses[i]
	int mWinsAgainstStandardAgents[][];
	
	// mWinsPerMap[a][o] = # games won by agent _a_ on map mBoardSets[o]
	int mWinsPerMap[][];
	
	// mRanking[r] = index of agent with tournament rank r
	int mRanking[];
	
	// mRankingPerAgent[a] = agent a's ranking
	int mRankingPerAgent[];
	
	// classes that make up each team for a given game
	Class mTeamClasses[];
	
	// vector of exceptions thrown by each team.
	// can be used to put stacktraces into each team's stats file
	Vector mTeamExceptions[];
	
	GameColorMap mColorMap = new DefaultGameColorMap();
	
	int mNumAgentsPerTeam = 2;
	
	
	public static void main( String [] argv ) {
		new Tournament();
		}
	
	public Tournament() {

		// first read settings from our ini file
		try {
			Reader reader =
				new BufferedReader(
					new InputStreamReader(
						new FileInputStream(
							mIniFileName ) ) );

			StreamTokenizer tokenizer = 
				new StreamTokenizer( reader );

		   
			tokenizer.eolIsSignificant( false );
			tokenizer.slashStarComments( true );
			tokenizer.slashSlashComments( true );


			// we use underscore in our keynames
			tokenizer.wordChars( '_', '_' );
			
			// tokenize until the end of the file
			while( tokenizer.nextToken() != StreamTokenizer.TT_EOF ) {
				String keyName = tokenizer.sval;

				// move on to the value for this key
				tokenizer.nextToken();

				if( keyName.equals( "correct_agent_code_points" ) ) {
					mCorrectAgentPoints = (int)( tokenizer.nval );
					}
				else if( keyName.equals( "exception_penalty" ) ) {
					mExceptionPenalty = (int)( tokenizer.nval );
					}
				else if( keyName.equals( "map_fraction_to_win" ) ) {
					mMapFractionToWin = (float)( tokenizer.nval );
					}
				else if( keyName.equals( "tournament_performance_points") ) {
					mTournamentPoints = (int)( tokenizer.nval );
					}
				else {
					System.out.println(
						"Unexpected token found in INI file at line "
						+ tokenizer.lineno() + ":  "
						+ keyName );
					}
				}
			// close the file
			reader.close();
			}
		catch( Exception inE ) {
			System.out.println(
				"Failed to read from INI file:  "
				+ mIniFileName);
			inE.printStackTrace();
			System.exit(1);	
			}	
		
		

		try {
			// load all file board sets
			File setFile = new File( mSetsDirName );
			String setFiles[] = setFile.list();

			// only consider directories
			Vector nonDirFiles = new Vector();
			for( int i=0; i<setFiles.length; i++ ) {
				File thisFile = new File( setFiles[i] );
				if( !thisFile.isDirectory() ) {
					nonDirFiles.addElement( setFiles[i] );
					}
				}
			
			mBoardSets = new FileSet[ nonDirFiles.size() ];
			mBoardSetNames = new String[ nonDirFiles.size() ];
			for( int i=0; i<nonDirFiles.size(); i++ ) {
				String fileName = mSetsDirName + File.separatorChar 
					+ (String)( nonDirFiles.elementAt( i ) );
				mBoardSets[i] =  new FileSet( new File( fileName ) );
				mBoardSetNames[i] =
					(String)( nonDirFiles.elementAt( i ) );
				}
			}
		catch( IOException inIOE ) {
			System.out.println( "Failed to load board sets from directory." );
			inIOE.printStackTrace();
			System.exit(1);
			}


		// read each standard class from file
		try {
			Reader reader =
				new BufferedReader(
					new InputStreamReader(
						new FileInputStream(
							mStandardAgentClassesFileName ) ) );

			StreamTokenizer tokenizer = 
				new StreamTokenizer( reader );
			
			tokenizer.eolIsSignificant( false );
			tokenizer.slashStarComments( true );
			tokenizer.slashSlashComments( true );

			tokenizer.nextToken();
			int numClasses = (int)( tokenizer.nval );

			mStandardAgentClasses = new Class[numClasses];
			mPointsPerStandardAgent = new int[numClasses];
			
			try {
				for( int i=0; i<numClasses; i++ ) {
					tokenizer.nextToken();
					String className = tokenizer.sval;
					mStandardAgentClasses[i] = Class.forName( className );

					// read the maximum point value
					tokenizer.nextToken();
					mPointsPerStandardAgent[i] =
						(int)( tokenizer.nval );
					}
				}
			catch( ClassNotFoundException inCNFE ) {
				System.out.println( "Class not found from " +
									mStandardAgentClassesFileName +
									"at line " +
									tokenizer.lineno() +
									": " + inCNFE.getMessage() );
				// close the file
				reader.close();
				throw new Exception();
				}
			
			// close the file
			reader.close();
			}
		catch( Exception inE ) {
			System.out.println( "Failed to load standard agent classes." );
			inE.printStackTrace();
			System.exit(1);	
			}	

		
				
		// read each competitor class from file
		try {
			Reader reader =
				new BufferedReader(
					new InputStreamReader(
						new FileInputStream(
							mCompetitorAgentClassesFileName ) ) );

			StreamTokenizer tokenizer = 
				new StreamTokenizer( reader );
			
			tokenizer.eolIsSignificant( false );
			tokenizer.slashStarComments( true );
			tokenizer.slashSlashComments( true );

			tokenizer.nextToken();
			int numClasses = (int)( tokenizer.nval );
			mAvailableCompetitorClasses = new Class[numClasses];
			mWinsAgainstStandardAgents =
				new int[numClasses][ mStandardAgentClasses.length ];
			mCompetitorWins = new int[numClasses];
			mWinsPerMap = new int[numClasses][mBoardSets.length];
			mTeamExceptions = new Vector[numClasses];
			try {
				for( int i=0; i<numClasses; i++ ) {
					tokenizer.nextToken();
					String className = tokenizer.sval;
					mAvailableCompetitorClasses[i] =
						Class.forName( className );

					for( int s=0; s<mStandardAgentClasses.length; s++ ) {
						mWinsAgainstStandardAgents[i][s] = 0;
						}
					
					mCompetitorWins[i] = 0;
					mTeamExceptions[i] = new Vector();
					for( int o=0; o<mBoardSets.length; o++ ) {
						mWinsPerMap[i][o] = 0;
						}
					}
				}
			catch( ClassNotFoundException inCNFE ) {
				System.out.println( "Class not found from " +
									mCompetitorAgentClassesFileName +
									"at line " +
									tokenizer.lineno() +
									": " + inCNFE.getMessage() );
				// close the file
				reader.close();
				throw new Exception();
				}
			
			// close the file
			reader.close();
			}
		catch( Exception inE ) {
			System.out.println( "Failed to load competitor agent classes." );
			inE.printStackTrace();
			System.exit(1);	
			}

		
		mTeamClasses = new Class[2];
		
	  		
		// for each competitor
		for( int a=0; a<mAvailableCompetitorClasses.length; a++ ) {
			System.out.println( "Playing team " + a + " against all." );
			
			mTeamClasses[0] = mAvailableCompetitorClasses[a];
			
			// for each obstacle set
			for( int o=0; o<mBoardSets.length; o++ ) {
				//System.out.println( "Obstacle set " + o );
				
				// play competitor against each standard agent
				for( int i=0; i<mStandardAgentClasses.length; i++ ) {
					//System.out.println( "Standard class opponent " + i );
					
					mTeamClasses[1] = mStandardAgentClasses[i];
					mGame = new Game( mTeamClasses, mColorMap, mBoardSets[o], 
						mBoardSets[o], mNumAgentsPerTeam, this );
					
					mGame.setDelay( 0 );
					int drawThreshold = mBoardSets[o].getBoardSize() * 
						mBoardSets[o].getBoardSize() * 2;
					int winner = mGame.playRound( drawThreshold, false );
					
					if( winner == 4 ) {
						// competitor threw an exception
						mTeamExceptions[a].addElement( 
							mGame.getTeamException() );
						System.out.println( "competitor " + a +
								" threw an exception." );	
						}
					if( winner == 3 ) {
						System.out.println(
							"Standard agent threw an exception!" );
						mGame.getTeamException().printStackTrace();
						System.exit(1);
						}
					else if( winner == 0 ) {
						// competitor won
						mWinsAgainstStandardAgents[a][i]++;
						mWinsPerMap[a][o]++;
						//System.out.println( "" + a + " won." );
						}
					else if( winner == 1 ) {
						//System.out.println( "Standard class won." );
						}
					else {
						//System.out.println( "Draw." );
						}
					}
					
				// play competitor against each other competing agent	
				for( int b=0; b<mAvailableCompetitorClasses.length; b++ ) {
					
					// don't play agent against itself	
					if( b!= a ) {	
						System.out.println( "Tournament opponent " + b );
						mTeamClasses[1] = mAvailableCompetitorClasses[b];
						mGame = new Game( mTeamClasses,
										  mColorMap,
										  mBoardSets[o], 
										  mBoardSets[o],
										  mNumAgentsPerTeam,
										  this );
						
						mGame.setDelay( 0 );
						int winner = mGame.playRound( mDrawThreshold, false );

						if( winner == 4 ) {
							// competitor a threw an exception
							mTeamExceptions[a].addElement( 
								mGame.getTeamException() );
							// competitor b won
							mCompetitorWins[b]++;
							mWinsPerMap[b][o]++;
							System.out.println( "competitor " + a +
								" threw an exception." );	
							}
						if( winner == 3 ) {
							// competitor b threw an exception
							mTeamExceptions[b].addElement( 
								mGame.getTeamException() );
							// competitor a won
							mCompetitorWins[a]++;
							mWinsPerMap[a][o]++;
							System.out.println( "competitor " + b +
								" threw an exception." );
							}
						else if( winner == 0 ) {
							// competitor a won
							mCompetitorWins[a]++;
							mWinsPerMap[a][o]++;
							//System.out.println( "" + a + " won." );
							}
						else if( winner == 1 ) {
							// competitor b won
							mCompetitorWins[b]++;
							mWinsPerMap[b][o]++;
							//System.out.println( "" + b + " won." );
							}
						else {
							//System.out.println( "Draw." );
							}	
						}
					}	// end for all other competitors	
				}	// end for all obstacle sets
			}	// end for all competitors
		
		
		System.out.println( "Sorting agents" );
		
		mRanking = new int[mAvailableCompetitorClasses.length];
		
		// start out unsorted
		for( int a=0; a<mAvailableCompetitorClasses.length; a++ ) {
			mRanking[a] = a;
			}
		
		// bubble sort
		for( int a=0; a<mAvailableCompetitorClasses.length; a++ ) {
			for( int b=a+1; b<mAvailableCompetitorClasses.length; b++ ) {
				if( mCompetitorWins[mRanking[a]] 
					< mCompetitorWins[mRanking[b]] ) {
					
					int temp = mRanking[a];
					mRanking[a] = mRanking[b];
					mRanking[b] = temp;
					}
				}
			}
		
		mRankingPerAgent = new int[mAvailableCompetitorClasses.length];
		int currentRanking = 1;
		int runningRanking = 1;
		for( int a=0; a<mAvailableCompetitorClasses.length; a++ ) {
			mRankingPerAgent[ mRanking[a] ] = currentRanking;
			if( a < mAvailableCompetitorClasses.length - 1
				&& mCompetitorWins[mRanking[a+1]] == 
				mCompetitorWins[mRanking[a]] ) {
				// don't increment current ranking, since tie
				}
			else {
				// increment current ranking
				currentRanking = runningRanking + 1;
				}
			runningRanking++;
			}
		
		System.out.println( "Writing Stats" );
		try {
			int maxNumPoints = 0;
			for( int s=0; s<mStandardAgentClasses.length; s++ ) {
				maxNumPoints += mPointsPerStandardAgent[s];
				}
			maxNumPoints +=
				mTournamentPoints
				+ mCorrectAgentPoints;
			
			for( int a=0; a<mAvailableCompetitorClasses.length; a++ ) {
				String fullClassName = mAvailableCompetitorClasses[a].getName();
				String partialName = fullClassName.substring(
					fullClassName.lastIndexOf('.') + 1 );
				String emailAddress = partialName.substring( 0,
					partialName.lastIndexOf('A') );	

				// put file into stats directory
				String fileName =
					mStatsDirName +
					File.separator +
					emailAddress;

				System.out.println( "Writing stats file: " + fileName );

				File statsFile = new File( fileName );
				FileOutputStream fileStream = new FileOutputStream( statsFile );
				PrintWriter fileWriter =  new PrintWriter( fileStream );
				
				fileWriter.println( "Subject: Stats for " + 
					mAvailableCompetitorClasses[a].getName() );
				fileWriter.println( "--- Stats for " + 
					mAvailableCompetitorClasses[a].getName() + " ---" );

				fileWriter.println();
				fileWriter.println( "--- Wins per map ---" );

				fileWriter.println( padString( "Map", 25 ) + "Wins on Map" );
				for( int o=0; o<mBoardSets.length; o++ ) {
					fileWriter.println( padString( mBoardSetNames[o], 25 ) + 
						+ mWinsPerMap[a][o] );
					}

				fileWriter.println();
				fileWriter.println( "--- Wins per opponent ---" );

				fileWriter.println( padString( "Opponent", 25 ) + 
					"Wins Against Opponent" );

				for( int i=0; i<mStandardAgentClasses.length; i++ ) {
					fileWriter.println( 
						padString( mStandardAgentClasses[i].getName(), 25 ) +
						+ mWinsAgainstStandardAgents[a][i] );
					}
				
				fileWriter.println();
				fileWriter.println( "--- Total tournament wins ---" );
				fileWriter.println( "" + mCompetitorWins[a] );
				
				fileWriter.println();
				fileWriter.println( "--- Tournament ranking ---" );
				fileWriter.println( "" + mRankingPerAgent[a] + " out of " +
					mAvailableCompetitorClasses.length );
				
				fileWriter.println();
				fileWriter.println( "--- Number of exceptions thrown ---" );
				fileWriter.println( mTeamExceptions[a].size() );
				fileWriter.println( "Stack traces for each exception are " +
					"included at the end of this message." );
				
				fileWriter.println();
				fileWriter.println( "--- Grading ---" );
				
				int agentCodeScore = mCorrectAgentPoints;
				
				if( mTeamExceptions[a].size() > 0 ) {
					agentCodeScore -= mExceptionPenalty;
					fileWriter.println(
						padString( "Exceptions thrown this round:", 30 ) +
						"-" + mExceptionPenalty );
					}
				
				fileWriter.println( padString( "Correct agent code:", 30 ) +
					agentCodeScore + "/" + mCorrectAgentPoints );

				int total = 0;
				
				// print score against each standard agent to file
				for( int s=0; s<mStandardAgentClasses.length; s++ ) {
					int pointsForAgent = 0;
					int requiredWins =
						(int)( mBoardSets.length * mMapFractionToWin );
					int maxPoints = mPointsPerStandardAgent[s];
					
					if( mWinsAgainstStandardAgents[a][0] >= requiredWins ) {
						pointsForAgent = maxPoints;
						}
					else {
						pointsForAgent = 
							(int)(
								( (float)mWinsAgainstStandardAgents[a][s] / 
								  (float)requiredWins ) *
								maxPoints );
						}		
					fileWriter.println(
						padString( "Performance vs. "
								   + mStandardAgentClasses[s].getName()
								   + ":",
								   50 ) +
						pointsForAgent + "/" + maxPoints );
					
					total += pointsForAgent;
					}

				
				int pointsForTournament = 
					(int)( mTournamentPoints *
						   ( ( mAvailableCompetitorClasses.length - 
							   ( mRankingPerAgent[a] - 1 ) ) /
							 (float)mAvailableCompetitorClasses.length ) );
				fileWriter.println(
					padString( "Performance in tournament:", 30 ) +
					pointsForTournament + "/" + mTournamentPoints );
				
				total += pointsForTournament +
					agentCodeScore;

				
				
				fileWriter.println( padString( "Total:", 30 ) +
					total + "/" + maxNumPoints );
				
				if( mTeamExceptions[a].size() > 0 ) {
					fileWriter.println();
					fileWriter.println( "--- Exception stack traces ---" );
					for( int e=0; e<mTeamExceptions[a].size(); e++ ) {
						fileWriter.println();
						Exception exception =
							(Exception)( mTeamExceptions[a].elementAt( e ) );
						exception.printStackTrace( fileWriter );	
						}
					}
				
				// for sendmail
				fileWriter.println(".");
				
				fileWriter.close();
				fileStream.close();
				}
			}
		catch( IOException inIOE ) {
			System.out.println( "Failed to write stats files." );
			inIOE.printStackTrace();
			}
		try {
			File statsFile = new File( mStatsDirName +
									   File.separator +
									   "ranking.txt" );
			FileOutputStream fileStream = new FileOutputStream( statsFile );
			PrintWriter fileWriter =  new PrintWriter( fileStream );
			fileWriter.println( "--- Final ranking ---" );
			fileWriter.println( padString( "Rank", 10 ) +
				padString( "Agent class", 25 ) + 
				"Wins" );
			for( int a=0; a<mAvailableCompetitorClasses.length; a++ ) {
				int agentIndex = mRanking[a];
				int rank = mRankingPerAgent[agentIndex];
				String fullClassName = 
					mAvailableCompetitorClasses[agentIndex].getName();
				String partialName = fullClassName.substring(
						fullClassName.lastIndexOf('.') + 1 );
				int wins = mCompetitorWins[agentIndex];
				fileWriter.println( padString( "" + rank, 10 ) +
					padString( partialName, 25 ) + wins );
				}
			fileWriter.close();
			fileStream.close();	
			}
		catch( IOException inIOE ) {
			System.out.println( "Failed to write final ranking file." );
			inIOE.printStackTrace();
			}
		
		// print to console also
		System.out.println( "--- Final ranking ---" );
		System.out.println( padString( "Rank", 10 ) +
			padString( "Agent class", 25 ) + 
			"Wins" );
		for( int a=0; a<mAvailableCompetitorClasses.length; a++ ) {
			int agentIndex = mRanking[a];
			int rank = mRankingPerAgent[agentIndex];
			String fullClassName = 
				mAvailableCompetitorClasses[agentIndex].getName();
			String partialName = fullClassName.substring(
					fullClassName.lastIndexOf('.') + 1 );
			int wins = mCompetitorWins[agentIndex];
			System.out.println( padString( "" + rank, 10 ) +
				padString( partialName, 25 ) + wins );
			}
			
		}	// end of Tournament constructor
	
	
	

/*	
	private class GameThread extends Thread {
		public void run() {
			String delayName = mDelayChoice.getSelectedItem();
			try {
				int delayInt = Integer.parseInt( delayName );
				mGame.setDelay( delayInt );
				}
			catch( NumberFormatException inNFE ) {
				}
			
			int winner = mGame.playRound( mDrawThreshold );
			while( winner != -1 ) {
				// game ended in a win... reset board and start again.
				mGame.setPause( true );
				scored( winner );
				mPauseButton.setLabel( "Un-Pause" );
				winner = mGame.playRound( mDrawThreshold );
				}
			// else game was canceled or was a draw.
			// do nothing but return	
			}
		}
	*/

	
	private String padString( String inString, int desiredLength ) {
		while( inString.length() < desiredLength ) {
			inString = inString + " ";
			}
		return inString;
		}
	
	public void stateChanged() {
		//System.out.println( "Step" );
		// do nothing
		}
	
	}
